import parse
from collections import OrderedDict
import mysql.connector
from jinja2 import Template
from tmg.data import logs


class Client:
    """
    Client to bundle MySQL functions.

    Args:
        connection_string (str): The MySQL connection string. For example: ``{my-username}:{my-password}@{mysql-host}:{mysql-port}/{my-database}``

    """

    def __init__(self, connection_string):

        parsed_connection = parse.parse('{username}:{password}@{host}:{port}/{database}', connection_string)

        self.connection = mysql.connector.connect(
            host=parsed_connection['host'],
            port=parsed_connection['port'],
            user=parsed_connection['username'],
            password=parsed_connection['password'],
            database=parsed_connection['database']
        )

    def get_table_schema(self, database, table):
        """ Get the table schema from MySQL database

        Args:
            database (str): The Database name
            table (str): The Table name

        Returns:
            dict: Schema. For example: ``{column_name: {'type':'some_type', 'default':'some_default_value'}}``
        """

        cursor = self.connection.cursor()
        logs.client.logger.info('Getting Columns from MySQL table {}.{}'.format(database, table))
        cursor.execute("SHOW columns from `{database}`.`{table}`".format(
            database=database, table=table
        ))

        fields = OrderedDict({
            column[0]: {"type": column[1], "default": column[4]} for column in cursor.fetchall()
        })

        cursor.close()

        return fields

    def upload_table(self, file_path, database, table):
        """Upload CSV file to MySQL from local path

        Args:
            file_path (str): The CSV file local path
            database (str): The database name
            table (str): The table name
        """

        # get table schema
        fields = self.get_table_schema(database, table)

        # make the load infile query
        load_data_infile_query = Template("""
            LOAD DATA LOCAL infile '{{ file_path }}'
            REPLACE
            INTO TABLE `{{database}}`.`{{table}}`
            CHARACTER SET 'utf8mb4'
            FIELDS TERMINATED BY ','
            OPTIONALLY ENCLOSED BY '\"'
            ESCAPED BY '\"'
            (
            {% for field in fields %}
                @{{field}}
                {% if not loop.last %}, {% endif %}
            {% endfor %}
            )
            SET
            {% for field in fields %}
            {{ field }} = nullif(@{{field}}, '')
            {% if not loop.last %},{% endif %}
            {% endfor %}
            ;
        """).render(file_path=file_path, database=database, table=table, fields=fields)

        logs.client.logger.info('Loading to MySQL table {}.{} from file {}'.format(database, table, file_path))
        cursor = self.connection.cursor()

        cursor.execute("ALTER TABLE `{}`.`{}` DISABLE KEYS;".format(database, table))
        cursor.execute("SET FOREIGN_KEY_CHECKS = 0;")
        cursor.execute("SET UNIQUE_CHECKS = 0;")
        cursor.execute(load_data_infile_query)
        cursor.execute("SET UNIQUE_CHECKS = 1;")
        cursor.execute("SET FOREIGN_KEY_CHECKS = 1;")
        cursor.execute("ALTER TABLE `{}`.`{}` ENABLE KEYS;".format(database, table))

        cursor.close()



import time
import uuid
import os
import boto3
import parse
import sys

from google.cloud import bigquery
from google.cloud import storage
from oauth2client.client import GoogleCredentials
from googleapiclient import discovery
from google.api_core import exceptions
from tmg.data import mysql
from tmg.data import ftp
from tmg.data._helper import merge_files
from tmg.data._helper import clean_mysql_export, get_bq_write_disposition
from tmg.data import logs
from tmg.data import salesforce
from tmg.data import gs
from tmg.data import s3
from tmg.data import bq


class Client:
    """
    Client to bundle transfers from a source to destination.

    Args:
        project (str): The Project ID for the project which the client acts on behalf of.
    """

    def __init__(self, project):
        self.project = project

    def bq_to_gs(self, table, bucket_name, separator=',', print_header=True, compress=False):
        """Extract BigQuery table into the GoogleStorage

        Args:
            table (str):  The BigQuery table name. For example: ``my-project-id.you-dataset.my-table``
            bucket_name (str):  The name of the bucket in GoogleStorage.
            separator (:obj:`str`, optional): The separator. Defaults to :data:`,`.
            print_header (:obj:`boolean`, optional):  True to print a header row in the exported data otherwise False. Defaults to :data:`True`.
            compress (:obj:`boolean`, optional): True to apply a GZIP compression. False to export without compression.

        Returns:
            list: The list of GoogleStorage paths for the uploaded files into the GoogleStorage.
            if the table is big, the exported files will be multiple.

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.bq_to_gs('my-project-id.some_dataset.some_table', 'some-bucket-name')
        """

        project, dataset_id, table_id = table.split('.')
        dataset_ref = bigquery.DatasetReference(
            project=project, dataset_id=dataset_id)
        table_ref = bigquery.TableReference(dataset_ref, table_id=table_id)

        bq_client = bigquery.Client(project=self.project)
        logs.client.logger.info('Extracting from {table} to gs://{bucket_name}/{table_id}_*'.format(
            bucket_name=bucket_name, table_id=table_id, table=table)
        )
        extract_job = bq_client.extract_table(
            source=table_ref,
            destination_uris='gs://{bucket_name}/{table_id}_*'.format(
                bucket_name=bucket_name, table_id=table_id),
            job_config=bigquery.ExtractJobConfig(
                field_delimiter=separator,
                print_header=print_header,
                compression=bigquery.Compression.GZIP if compress else None
            )
        )
        extract_job.result()

        storage_client = storage.Client(project=self.project)
        logs.client.logger.info(
            'Getting the list of available blobs in gs://{}'.format(bucket_name))
        blobs = storage_client.list_blobs(bucket_name)

        return ['gs://{}/{}'.format(bucket_name, blob.name) for blob in blobs]

    def gs_to_bq(self, gs_uris, table, write_preference, auto_detect=True, skip_leading_rows=True, separator=',',
                 schema=(), partition_date=None, max_bad_records=0):
        """Load file from Google Storage into the BigQuery table

        Args:
            gs_uris (Union[str, Sequence[str]]):  The Google Storage uri(s) for the file(s). For example: A single file: ``gs://my_bucket_name/my_filename``, multiple files: ``[gs://my_bucket_name/my_first_file, gs://my_bucket_name/my_second_file]``.
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            write_preference (str): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
            auto_detect (boolean, Optional):  True if the schema should automatically be detected otherwise False. Defaults to :data:`True`.
            skip_leading_rows (boolean, Optional):  True to skip the first row of the file otherwise False. Defaults to :data:`True`.
            separator (str, Optional): The separator. Defaults to :data:`,`
            schema (tuple): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field', 'STRING'))``
            partition_date (str, Optional): The ingestion date for partitioned BigQuery table. For example: ``20210101``. The partition field name will be __PARTITIONTIME.
            max_bad_records (int, Optional): The maximum number of rows with errors. Defaults to :data:0

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.gs_to_bq(gs_uris='gs://my-bucket-name/my-filename',table='my-project-id.my_dataset.my_table')
        """

        project, dataset_id, table_id = table.split('.')
        dataset_ref = bigquery.DatasetReference(
            project=project, dataset_id=dataset_id)

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.CSV,
            skip_leading_rows=1 if skip_leading_rows else 0,
            autodetect=auto_detect,
            field_delimiter=separator,
            write_disposition=get_bq_write_disposition(write_preference),
            allow_quoted_newlines=True,
            max_bad_records=max_bad_records
        )

        if partition_date:
            job_config.time_partitioning = bigquery.TimePartitioning(
                type_=bigquery.TimePartitioningType.DAY)
            table_id += '${}'.format(partition_date)

        table_ref = bigquery.TableReference(dataset_ref, table_id=table_id)

        bq_client = bq.Client(project)
        try:
            bq_client.create_dataset(dataset_id)
        except exceptions.Conflict:
            logs.client.logger.info(
                'Dataset {} Already exists'.format(dataset_id))

        if schema:
            job_config.schema = [bigquery.SchemaField(
                schema_field[0], schema_field[1]) for schema_field in schema]

        bigquery_client = bigquery.Client(project=self.project)
        logs.client.logger.info(
            'Loading BigQuery table {} from {}'.format(table, gs_uris))
        job = bigquery_client.load_table_from_uri(
            gs_uris, table_ref, job_config=job_config)

        job.result()

    def bq_to_mysql(self, connection_string, bq_table, mysql_table):
        """Export from BigQuery table to MySQL table

        .. note:: The CloudSQL service account which can be found in Cloud SQL UI
              needs to have a storage object admin access.

        Args:
            connection_string (str): The MySQL connection string. For example: ``{my-username}:{my-password}@{mysql-host}:{mysql-port}/{my-database}``
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            mysql_table (str):  The MySQL table. For example: ``my-mysql-database.my-table``

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.bq_to_mysql(
            >>>     connection_string='{my-username}:{my-password}@{mysql-host}:{mysql-port}/{my-database}'
            >>>     bq_table='my-project-id.my-dataset.my-table'
            >>>     mysql_table='my-mysql-database.my-mysql-table'
            >>> )
        """

        # download the files locally
        bq_client = bq.Client(self.project)
        file_names = bq_client.download_table(bq_table, print_header=False)

        # merge the downloaded files
        if len(file_names) > 1:
            logs.client.logger.info('Merging {}'.format(','.join(file_names)))
            output_file_name = merge_files(file_names)
        else:
            output_file_name = file_names[0]

        # upload the merged file into mysql
        database, table = mysql_table.split(".")
        mysql_client = mysql.Client(connection_string)
        mysql_client.upload_table(
            file_path=output_file_name, database=database, table=table)

    def mysql_to_gs(self, instance_name, database, query, gs_uri):
        """Export from MySQL to Google Storage in CSV format

        .. note:: Be aware that the exported CSV file has "N as a field value when there is no value for that field.

        Args:
            instance_name (str): The CloudSQL instance name. It's the instance name in CloudSQL UI but without project.location
            database (str): The MySQL database name
            query (str): The query to get the data from MySQL
            gs_uri (str): The GoogleStorage uri. For example: ``gs://bucket_name/file_name``

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.mysql_to_gs(
            >>>     instance_name='{my-mysql-instance-name}',
            >>>     database='my-database',
            >>>     query='SELECT * FROM my-table',
            >>>     gs_uri='gs://my-bucket-name/my-filename'
            >>> )
        """

        # make the discovery service
        credentials = GoogleCredentials.get_application_default()
        service = discovery.build(
            'sqladmin', 'v1beta4', credentials=credentials)

        # make the request body
        instances_export_request_body = {
            "exportContext": {
                "fileType": "CSV",
                "uri": gs_uri,
                "databases": [database],
                "csvExportOptions": {
                    "selectQuery": query
                }
            }
        }

        # send the request and get the operation id
        logs.client.logger.info(
            'Exporting from MySQL database {} with query "{}" to {}'.format(database, query, gs_uri))
        request = service.instances().export(
            project=self.project,
            instance=instance_name,
            body=instances_export_request_body
        )
        response = request.execute()
        operation_id = response['name']

        # wait until the job is done
        status = 'PENDING'
        while status in ['PENDING', 'RUNNING']:
            request = service.operations().get(project=self.project, operation=operation_id)
            status = request.execute()['status']
            # Avoid to hammer the APIs (100queries per users every 100seconds is the maximum).
            time.sleep(2)

        if status != 'DONE':
            raise Exception(
                'Failed to export data from MySQL, process status {}'.format(status))

    def mysql_to_bq(self, instance_name, database, query, bq_table, bq_table_schema, write_preference, partition_date=None):
        """Export from MySQL to BigQuery

        Args:
            instance_name (str): The CloudSQL instance name. It's the instance name in CloudSQL UI but without project.location
            database (str):  The MySQL database name
            query (str): The query to get the data from MySQL
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            bq_table_schema (tuple): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field', 'STRING'))``
            write_preference (str): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
            partition_date (str, Optinal): The ingestion date for partitioned BigQuery table. For example: ``20210101``. The partition field name will be __PARTITIONTIME


        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.mysql_to_bq(
            >>>     instance_name='{my-mysql-instance-name}',
            >>>     database='my-database',
            >>>     query='SELECT * FROM my-table',
            >>>     bq_table='my-project-id.my-dataset.my-table',
            >>>     bq_table_schema=(('first_field','STRING'),('second_field', 'STRING'))
            >>> )

        """

        # make the temp bucket
        storage_client = storage.Client(project=self.project)
        tmp_bucket_name = str(uuid.uuid4())
        logs.client.logger.info(
            'Creating the temporary bucket {}'.format(tmp_bucket_name))
        tmp_bucket = storage_client.create_bucket(
            tmp_bucket_name, location='EU')
        tmp_file_name = str(uuid.uuid4())

        # export from mysql to temp bucket
        self.mysql_to_gs(
            instance_name=instance_name,
            database=database,
            query=query,
            gs_uri='gs://{}/{}.csv'.format(tmp_bucket_name, tmp_file_name)
        )

        # download the exported file from the temp bucket into local
        blob = tmp_bucket.blob('{}.csv'.format(tmp_file_name))
        logs.client.logger.info('Downloading {}'.format(blob.name))
        blob.download_to_filename('{}.csv'.format(tmp_file_name))

        # clean the file by replacing ,"N with space
        logs.client.logger.info(
            'Cleaning the exported file {}.csv'.format(tmp_file_name))
        cleaned_file_name = clean_mysql_export('{}.csv'.format(tmp_file_name))

        # upload the cleaned file into BigQuery
        bq_client = bq.Client(project=self.project)
        bq_client.upload_table(
            file_path=cleaned_file_name,
            table=bq_table,
            auto_detect=False,
            schema=bq_table_schema,
            skip_leading_rows=False,
            write_preference=write_preference,
            partition_date=partition_date
        )

        # cleanup
        blob.delete()
        logs.client.logger.info(
            'Deleting temporary bucket {}'.format(tmp_bucket_name))
        tmp_bucket.delete()

    def ftp_to_bq(self, ftp_connection_string, ftp_filepath, bq_table, write_preference, separator=',',
                  skip_leading_rows=True, bq_table_schema=[], partition_date=None):
        """Export from FTP to BigQuery

        Args:
            ftp_connection_string (str): The FTP connection string in the format {username}:{password}@{host}:{port}
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            write_preference (str): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
            ftp_filepath (str): The path to the file to download.
            separator (:obj:`str`, Optional): The separator. Defaults to :data:`,`.
            skip_leading_rows (boolean, Optional):  True to skip the first row of the file otherwise False. Defaults to :data:`True`.
            bq_table_schema (tuple, Optional): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field','STRING'))``
            partition_date (str, Optional): The ingestion date for partitioned BigQuery table. For example: ``20210101``. The partition field name will be __PARTITIONTIME

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.ftp_to_bq(
            >>>     ftp_connection_string='username:password@hots:port',
            >>>     ftp_filepath='/my-path/to-the-ftp-file',
            >>>     bq_table='my-project-id.my-dataset.my-table'
            >>> )

        """

        # download the ftp file
        ftp_client = ftp.Client(connection_string=ftp_connection_string)
        local_file = ftp_client.download_file(ftp_filepath)

        # upload the ftp file into BigQuery
        bq_client = bq.Client(project=self.project)
        bq_client.upload_table(
            file_path=local_file,
            table=bq_table,
            separator=separator,
            skip_leading_rows=skip_leading_rows,
            write_preference=write_preference,
            schema=bq_table_schema,
            partition_date=partition_date
        )

    def bq_to_ftp(self, bq_table, ftp_connection_string, ftp_filepath, separator=',', print_header=True):
        """Export from BigQuery to FTP

        Args:
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            ftp_connection_string (str): The FTP connection string in the format {username}:{password}@{host}:{port}
            ftp_filepath (str): The path to the file to download.
            separator (:obj:`str`, optional): The separator. Defaults to :data:`,`.
            print_header (boolean, Optional):  True to write header for the CSV file, otherwise False. Defaults to :data:`True`.

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.bq_to_ftp(
            >>>     bq_table='my-project-id.my-dataset.my-table',
            >>>     ftp_connection_string='username:password@hots:port',
            >>>     ftp_filepath='/my-path/to-the-ftp-file'
            >>> )

        """
        # download the the BigQuery table into local
        bq_client = bq.Client(project=self.project)
        local_files = bq_client.download_table(
            table=bq_table,
            separator=separator,
            print_header=print_header
        )

        # merge the files if they are more than one
        if len(local_files) > 1:
            logs.client.logger.info('Merging {}'.format(','.join(local_files)))
            merged_file = merge_files(local_files)
        else:
            merged_file = local_files[0]

        # upload the merged file
        ftp_client = ftp.Client(connection_string=ftp_connection_string)
        ftp_client.upload_file(local_path=merged_file,
                               remote_path=ftp_filepath)

    def salesforce_to_bq(self, salesforce_connection_string, salesforce_table, salesforce_columns, bq_table, bq_table_schema,
                         write_preference, salesforce_condition=None, salesforce_include_deleted=False,
                         fields_transform=[], partition_date=None):
        """Export from Salesforce to BigQuery

        Args:
            salesforce_connection_string (str): The Salesforce connection string in the format {username}:{password}:{token}@{domain}
                                                Domain should be either ``'login'`` for production server or ``'test'`` for UAT servers
            salesforce_table (str):  Saleforce Table name. For example: ``Account``.
            salesforce_columns (tuple): Salesforce Table columns. For example: ``('Id', 'FirstName', 'LastName')``.
            salesforce_condition (:obj:`str`, Optional): The condition which should apply to the table. For example: ``ModifiedDate > 2020-01-01``. Defaults to :data:`None`
            salesforce_include_deleted (:obj:`boolean`, Optional): Include deleted records in the returned result.
                                                                   IsDeleted field is available for each record. Defaults to :data:`False`

            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            bq_table_schema (tuple): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field','STRING'))``
            write_preference (str): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
            fields_transform(:obj:`list`, Optional): List of transformation functions per field. For example: ``[('FirstName', lambda name: name.lower())]``
            partition_date (str, Optional): The ingestion date for partitioned BigQuery table. For example: ``20210101``. The partition field name will be __PARTITIONTIME

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.salesforce_to_bq(
            >>>     salesforce_connection_string='username:password:token@domain',
            >>>     salesforce_table='Account'
            >>>     salesforce_columns=('Id', 'FirstName', 'LastName')
            >>>     bq_table='my-project-id.my-dataset.account'
            >>>     bq_table_schema=(('id','STRING'),('first_name','STRING'),('last_name', 'STRING'))
            >>> )

        """

        # download the Salesforce table into a local file
        salesforce_client = salesforce.Client(
            connection_string=salesforce_connection_string)
        local_file = salesforce_client.download_table(
            table=salesforce_table,
            columns=salesforce_columns,
            condition=salesforce_condition,
            include_deleted=salesforce_include_deleted,
            fields_transform=fields_transform
        )

        # upload the downloaded file into BigQuery
        bq_client = bq.Client(project=self.project)
        bq_client.upload_table(
            file_path=local_file,
            table=bq_table,
            write_preference=write_preference,
            schema=bq_table_schema,
            partition_date=partition_date
        )

    def gs_to_s3(self, gs_uri, s3_connection_string, s3_bucket):
        """
        Exports file from Google storage bucket to S3 bucket

        Args:
        gs_uri (str): Google storage uri path
        s3_connection_string (str): The S3 connection string in the format
                                    {region}:{access_key}:{secret_key}
        s3_bucket (str): s3 bucket name

        Example:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.gs_to_s3(gs_uri='gs://my-bucket-name/my-filename',
            >>>                 s3_connection_string='region:access_key:secret_key',
            >>>                 s3_bucket='bucket_name')
        """

        local_file = os.path.basename(gs_uri)
        gs_client = gs.Client(self.project)
        gs_client.download(gs_uri, local_file)

        s3_client = s3.Client(s3_connection_string)
        s3_client.upload(local_file,
                         s3_bucket)

    def s3_to_gs(self, s3_connection_string, s3_bucket_name,
                 s3_object_name, gs_bucket_name, gs_file_name=None):
        """
        Exports file(s) from S3 bucket to Google storage bucket

        Args:
          s3_connection_string (str): The S3 connection string in the format
                                   {region}:{access_key}:{secret_key}
          s3_bucket_name (str): s3 bucket name
          s3_object_name (str): s3 object name or prefix to match multiple files to copy
          gs_bucket_name (str): Google storage bucket name
          gs_file_name (str): GS file name

        Example:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.s3_to_gs(s3_connection_string='region:access_key:secret_key',
            >>>                 s3_bucket_name='my-s3-bucket_name',
            >>>                 s3_object_name='my-s3-file-prefix',
            >>>                 gs_bucket_name='my-gs-bucket-name',
            >>>                 gs_file_name='gs_file_name')
        """

        parsed_connection = parse.parse(
            '{region}:{access_key}:{secret_key}', s3_connection_string)

        # Retrieve the file(s) from S3 matching to the object
        logs.client.logger.info('Finding files in S3 bucket')
        s3_client = boto3.client('s3',
                                 region_name=parsed_connection['region'],
                                 aws_access_key_id=parsed_connection['access_key'],
                                 aws_secret_access_key=parsed_connection['secret_key'])
        response = s3_client.list_objects_v2(
            Bucket=s3_bucket_name, Prefix=s3_object_name)
        s3_files = []
        for content in response.get('Contents', []):
            s3_files.append(content.get('Key'))
        logs.client.logger.info(f'Found {str(s3_files)} files in S3')

        s3_client = s3.Client(s3_connection_string)
        gs_client = gs.Client(self.project)
        for s3_file in s3_files:
            s3_client.download(s3_bucket_name, s3_file)

            gs_file_name = (gs_file_name if gs_file_name is not None else s3_file) \
                if len(s3_files) == 1 else s3_file
            gs_client.upload(os.path.basename(s3_file),
                             gs_bucket_name, gs_file_name)

    def s3_to_bq(self, s3_connection_string, bucket_name, object_name,
                 bq_table, write_preference, auto_detect=True, separator=',',
                 skip_leading_rows=True, schema=[], partition_date=None):
        """
        Exports S3 file to BigQuery table

        Args:
          s3_connection_string (str): The S3 connection string in the format
                                      {region}:{access_key}:{secret_key}
          bucket_name (str): s3 bucket name
          object_name (str): s3 object name to copy
          bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
          write_preference (str): The option to specify what action to take when you load data from a source file. Value can be on of
                                            ``'empty'``: Writes the data only if the table is empty.
                                            ``'append'``: Appends the data to the end of the table.
                                            ``'truncate'``: Erases all existing data in a table before writing the new data.
          auto_detect (boolean, Optional):  True if the schema should automatically be detected otherwise False. Defaults to `True`.
          separator (str, optional): The separator. Defaults to `,`.
          skip_leading_rows (boolean, Optional):  True to skip the first row of the file otherwise False. Defaults to `True`.
          schema (list of tuples, optional): The BigQuery table schema. For example: ``[('first_field','STRING'),('second_field', 'STRING')]``
          partition_date (str, Optional): The ingestion date for partitioned BigQuery table. For example: ``20210101``. The partition field name will be __PARTITIONTIME

        Example:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.s3_to_bq(s3_connection_string='region:access_key:secret_key',
            >>>                 bucket_name='my-s3-bucket_name',
            >>>                 object_name='my-s3-object-name',
            >>>                 bq_table='my-project-id.my-dataset.my-table')
        """

        # Download S3 file to local
        s3_client = s3.Client(s3_connection_string)
        s3_client.download(bucket_name, object_name,
                           os.path.join('/tmp/', object_name))

        logs.client.logger.info('Loading S3 file to BigQuery table')
        bq_client = bq.Client(bq_table.split('.')[0])
        bq_client.upload_table(
            file_path=os.path.join('/tmp/', object_name),
            table=bq_table,
            write_preference=write_preference,
            separator=separator,
            auto_detect=auto_detect,
            skip_leading_rows=skip_leading_rows,
            schema=schema,
            partition_date=partition_date
        )
        logs.client.logger.info('Loading completed')

    def bq_to_salesforce(self, query: str, salesforce_connection_string: str, salesforce_table: str,
                         salesforce_bq_mapping: dict, time_field: dict = None) -> str:
        """Export from BigQuery to Salesforce

       Args:
           query (str): The Query to be sent to BigQuery. For example: `` Select * from project.database.table ``
           salesforce_connection_string (str): The Salesforce connection string in the format {username}:{password}:{token}@{domain}
                                               Domain should be either ``'login'`` for production server or ``'test'`` for UAT servers
           salesforce_table (str):  Saleforce Table name. For example: ``Account``.
           salesforce_bq_mapping (dict): Salesforce columns that map to BigQuery fields. For example: ``{'Id': 'id', 'FirstName':'first_name', 'LastName':'last_name'}``
                                           The Salesforce column should be the key and the Big Query field should be the value.
           time_field (:obj:`dict`, optional): Any timefields from BigQuery that will need to be converted to strings.
                                               The field name and desired format must be provided. For example: ``{"DateField": "%Y-%m-%dT%H:%M:%S"}``

       Examples:
           >>> from tmg.data import transfer
           >>> client = transfer.Client(project='my-project-id')
           >>> client.bq_to_salesforce(
           >>>     query='SELECT * FROM TABLE'    
           >>>     salesforce_connection_string='username:password:token@domain',
           >>>     salesforce_table='Account'
           >>>     salesforce_bq_mapping={'Id': 'id', 'FirstName':'first_name', 'LastName':'last_name'}
           >>>     time_fields={"DateField": "%Y-%m-%dT%H:%M:%S"}
           >>> )

        """
        # getting data from bq
        bq_client = bq.Client(project=self.project)
        bq_data = bq_client.run_query(query=query)
        if bq_data.total_rows == 0:
            logs.client.logger.error(
                "No data returned from bq please check your query")
            sys.exit(1)

        # building data list to be send to SF from BQ using field mapping
        sf_fields = salesforce_bq_mapping.keys()
        data = []
        for row in bq_data:
            sf_row = {}
            for field in sf_fields:
                try:
                    sf_row[field] = row[salesforce_bq_mapping[field]]
                    # udating time fields according for the format string
                    if time_field:
                        for tf, fs in time_field.items():
                            if field == tf:
                                sf_row[tf] = sf_row[tf].strftime(
                                    fs)
                except KeyError:
                    logs.client.logger.error(
                        f"Big Query row does not contain field {field}")
                    sys.exit(1)
            data.append(sf_row)
        # to check the data comming out of bq
        for i in range(len(data[:3])):
            logs.client.logger.info(
                f"showing line {i} of bq data \n {data[i]}")
        # sending data to salesforce via tmg salesforce client lib
        sf_client = salesforce.Client(salesforce_connection_string)
        update = sf_client.update_salesforce(salesforce_table, data)

        return update



import unittest
import os
import boto3

from tmg.data.s3 import Client
from tests.setup import setup

def setUpModule():
    setup.upload_s3_files()


def tearDownModule():
    setup.remove_s3_files()


class TestS3(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestS3, self).__init__(*args, **kwargs)
        self.setup = setup

    def test_download(self):
        test_client = Client(self.setup.s3_region,
                             self.setup.s3_access_key,
                             self.setup.s3_secret_key)

        test_client.download(setup.s3_bucket,
                             'download_sample.csv')
        self.assertTrue(os.path.exists('data/download_sample.csv'))

        test_client.download(setup.s3_bucket,
                             'download_sample.csv',
                             'data/download_s3_sample.csv')

        self.assertTrue(os.path.exists('data/download_s3_sample.csv'))

    def test_upload(self):
        test_client = Client(self.setup.s3_region,
                             self.setup.s3_access_key,
                             self.setup.s3_secret_key)

        test_client.upload('data/sample.csv',
                           setup.s3_bucket)
        s3_client = boto3.resource(service_name='s3',
                                   region_name=self.setup.s3_region,
                                   aws_access_key_id=self.setup.s3_access_key,
                                   aws_secret_access_key=self.setup.s3_secret_key)
        bucket = s3_client.Bucket(setup.s3_bucket)
        obj = list(bucket.objects.filter(Prefix='sample.csv'))
        self.assertTrue(any([w.key == 'sample.csv' for w in obj]))

        test_client.upload('data/s3_upload_file.csv',
                           setup.s3_bucket,
                           's3_upload_file.csv')
        s3_client = boto3.resource(service_name='s3',
                                   region_name=self.setup.s3_region,
                                   aws_access_key_id=self.setup.s3_access_key,
                                   aws_secret_access_key=self.setup.s3_secret_key)
        bucket = s3_client.Bucket(setup.s3_bucket)
        obj = list(bucket.objects.filter(Prefix='s3_upload_file.csv'))
        self.assertTrue(any([w.key == 's3_upload_file.csv' for w in obj]))




import delegator
import os

from google.cloud import bigquery


def merge_files(files_path, output_file_path=None):

    output = 'merged.csv'

    with open(output_file_path if output_file_path else output, 'w') as outfile:
        for file_path in files_path:
            with open(file_path) as infile:
                outfile.write(infile.read())

    return output


def clean_mysql_export(file_path):
    """Cleans MySQL export (CSV) format removing "N values and replacing them with null.
    Args:
        file_path (str): CSV file path

    Returns:
        str: Cleaned CSV file name
    """

    # Crazy SED command to clean Cloud SQL CSV export
    # that doesn't support Null values which are written as "N....seriously.
    command = "cat {file_path} | sed 's/,\"N,/,,/g' | sed 's/,\"N,/,,/g' | sed 's/^\"N,/,/g' | sed 's/,\"N$/,/g'" \
              " >> {file_name}_cleaned.csv"
    file_name = os.path.splitext(file_path)[0]
    command = command.format(file_path=file_path, file_name=file_name)
    delegator.run(command, block=True)

    return "{}_cleaned.csv".format(file_name)


def get_bq_write_disposition(write_preference):
    """Convert write_preference string to BigQuery WriteDisposition values

    Args:
        write_preference (str): The write preference string which should be 'truncate', 'append' or 'empty'

    Returns:
        bigquery.WriteDisposition: The BigQuery WriteDisposition value

    """

    disposition = {
        'truncate': bigquery.WriteDisposition.WRITE_TRUNCATE,
        'append': bigquery.WriteDisposition.WRITE_APPEND,
        'empty': bigquery.WriteDisposition.WRITE_EMPTY
    }

    return disposition.get(write_preference, None)



import boto3
import botocore
import os
import parse
import sys
from tmg.data import logs


class Client:
    """
        Client to s3 Storage functionality.

        Args:
            connection_string (str): The S3 connection string in the format
                                     {region}:{access_key}:{secret_key}
    """

    def __init__(self, connection_string):
        parsed_connection = parse.parse('{region}:{access_key}:{secret_key}', connection_string)
        self.s3_client = boto3.resource(service_name='s3',
                                        region_name=parsed_connection['region'],
                                        aws_access_key_id=parsed_connection['access_key'],
                                        aws_secret_access_key=parsed_connection['secret_key'])

    def download(self, bucket_name, object_name, local_path='.'):
        """
        Downloads a file from the s3 to the local system.

        Args:
            bucket_name (str): s3 bucket name
            object_name (str): s3 file name to download
            local_path (str): local file name with path. If local path is a directory
                              object name is used as local file name

        Example:
            >>> from tmg.data import s3
            >>> client = s3.Client('region:access_key:secret_key')
            >>> client.download(bucket_name='my-s3-bucket-name',
            >>>                 object_name='folder-name/object-name',
            >>>                 local_path='/my-local/folder/file.csv')
        """
        if os.path.isdir(local_path):
            filename = os.path.basename(object_name)
            local_path = os.path.join(local_path, filename)

        try:
            logs.client.logger.info(f"Starting {os.path.basename(object_name)} download from {bucket_name} s3 bucket")
            bucket = self.s3_client.Bucket(bucket_name)
            bucket.download_file(object_name, local_path)
        except botocore.exceptions.ClientError as e:
            logs.client.logger.error(f"The file {os.path.basename(object_name)} doesn't exists in the bucket, Terminating")
            sys.exit(1)

        logs.client.logger.info("File download completed")

    def upload(self, local_path, bucket_name, object_name=None):
        """
        Uploads a local file to s3 bucket

        Args:
            local_path (str): File with path to upload
            bucket_name (str): s3 bucket name
            object_name (str): S3 object name. If not specified then local file_name is used

        Example:
            >>> from tmg.data import s3
            >>> client = s3.Client('region:access_key:secret_key')
            >>> client.upload(local_path='/my-local/folder/file.csv',
            >>>               bucket_name='my-s3-bucket-name',
            >>>               object_name='object-name')
        """

        if object_name is None:
            object_name = os.path.basename(local_path)

        logs.client.logger.info(f"Starting {os.path.basename(local_path)} upload to {bucket_name} s3 bucket")
        bucket = self.s3_client.Bucket(bucket_name)
        bucket.upload_file(local_path, object_name)
        logs.client.logger.info("File upload completed")



import parse
import csv
import os
import sys
from time import sleep

from simple_salesforce import Salesforce, exceptions
from tmg.data import logs


class Client:
    """
    Client to bundle Salesforce functionality.

    Args:
        connection_string (str): The Salesforce connection string in the format {username}:{password}:{token}@{domain}
                                 Domain should be either ``'login'`` for production server or ``'test'`` for UAT servers

    """

    MAX_RETRIES = 10
    WAIT_BETWEEN_RETRY = 10

    def __init__(self, connection_string):

        parsed_connection = parse.parse(
            '{username}:{password}:{token}@{domain}', connection_string)
        self.sf = Salesforce(
            username=parsed_connection['username'],
            password=parsed_connection['password'],
            security_token=parsed_connection['token'],
            domain=parsed_connection['domain']
        )
        self.domain = parsed_connection['domain']

    def run_query(self, query, include_deleted=False):
        """Run the query on Salesforce server and return the result

        Args:
            query (str): The query string.
            include_deleted (:obj:`boolean`, Optional): Include deleted records in the returned result.
                                                        IsDeleted field is available for each record. Defaults to :data:`False`
        Returns:
            The query results

        """

        logs.client.logger.info(
            'Running query on Salesforce: {}'.format(query))
        return self.sf.query_all_iter(query, include_deleted=include_deleted)

    def download_table(self, table, columns, condition=None, include_deleted=False, local_folder='.', separator=',',
                       print_header=True, fields_transform=[]):
        """
        Export the table to the local file in CSV format.

        Args:
            table (str):  Table name. For example: ``Account``.
            columns (tuple): Table columns. For example: ``('Id', 'FirstName', 'LastName')``.
            condition (:obj:`str`, Optional): The condition which should apply to the table. For example: ``ModifiedDate > 2020-01-01``. Defaults to :data:`None`
            include_deleted (:obj:`boolean`, Optional): Include deleted records in the returned result.
                                                        IsDeleted field is available for each record. Defaults to :data:`False`
            local_folder (:obj:`str`, Optional):  The local folder with out the slash at end. For example: ``/some_path/some_inside_path``. Defaults to current path :data:`.`
            separator (:obj:`str`, Optional): The separator. Defaults to :data:`,`
            print_header (:obj:`boolean`, Optional):  True to print a header row in the exported file otherwise False. Defaults to :data:`True`.
            fields_transform(:obj:`list`, Optional): List of transformation functions per field. For example: ``[('FirstName', lambda name: name.lower())]``

        Returns:
            str: The output file path

        Examples:
            >>> from tmg.data import salesforce
            >>> client = salesforce.Client(username='username',password='password',token='token')
            >>> client.download_table(table='Account',columns=('Id', 'FirstName', 'LastName'))
        """

        query_str = 'SELECT {columns} FROM {table} {where_clause}'.format(
            columns=','.join(columns),
            table=table,
            where_clause='WHERE {}'.format(condition) if condition else ''

        )
        rows = self.run_query(query_str, include_deleted)

        output_file = '{}/{}.csv'.format(local_folder, table)
        logs.client.logger.info('Exporting data into {}'.format(output_file))

        csv_file = open(output_file, 'w')
        writer = csv.DictWriter(
            csv_file, fieldnames=columns, delimiter=separator)
        if print_header:
            writer.writeheader()

        # it's best to handle the exception inside the generator query_all_iter
        # but because the generator handled by simple_salesforce library it's not possible to do it.
        # the workaround is to handle the exception while using the generator. if the exception happens
        # then the generator starts from beginning. Unfortunately there is no way around it.
        try_count = 0
        while True:
            try:
                row = next(rows)  # get the next row from generator
            except StopIteration:  # end of iteration so break!
                break
            except (exceptions.SalesforceExpiredSession,
                    exceptions.SalesforceMalformedRequest,
                    exceptions.SalesforceAuthenticationFailed) as e:

                if try_count == self.MAX_RETRIES:
                    raise Exception(
                        "Reading from Salesforce failed. Reached maximum retries.")

                logs.client.logger.warning(
                    "Reading from Salesforce failed with this message: {message}."
                    "Sleeping for {wait_time} seconds before next try".format(
                        message=e.content[0]["message"],
                        wait_time=self.WAIT_BETWEEN_RETRY
                    )
                )

                sleep(self.WAIT_BETWEEN_RETRY)
                try_count += 1

                # cleanup and reset the generator
                os.remove(output_file)
                csv_file = open(output_file, 'w')
                writer = csv.DictWriter(
                    csv_file, fieldnames=columns, delimiter=separator)
                if print_header:
                    writer.writeheader()
                rows = self.run_query(query_str, include_deleted)

            else:
                row.pop('attributes')
                for field_transform in fields_transform:
                    row[field_transform[0]] = field_transform[1](
                        row[field_transform[0]])
                writer.writerow(row)

        csv_file.close()

        return output_file

    def update_salesforce(self, sf_table: str,  data: list) -> str:
        '''
        Bulk updates a table in salesforce
        Args:
            sf_table (str):  Table name. For example: ``Account``.
            data (list): List data to be sent to salesforce. For example: ``[{"FirstName": "abcd", "LastName":"xyz"}, {...}]``

        Returns:
            str: number of records updated sucessfully
        Raises: General Salesforce Error

        Examples:
            >>> from tmg.data import salesforce
            >>> client = salesforce.Client("username:password:token@domain")
            >>> client.update_salesforce(sf_table='Account',data=[{"FirstName": "abcd", "LastName":"xyz"}, {...})
        '''
        logs.client.logger.info(
            f"Updating {sf_table} table in Salesforce domain {self.domain}")
        # creating a SFBulkType object
        sf_table_object = self.sf.bulk.sf_table
        # subtituting sf_table into object_name variable
        setattr(sf_table_object, 'object_name', sf_table)
        # most errors are handled in the salesforce library
        update = sf_table_object.update(data)

        for row in update:
            if not row['success']:
                logs.client.logger.error(
                    f"Update failed with the following error {row['errors']}")
                sys.exit(1)
        success_ids = [x['id'] for x in update]
        logs.client.logger.info(
            f"Update successful : {len(success_ids)} records updated in table {sf_table} in {self.domain}")
        return "Success"



import uuid
import csv

from jinja2 import Template
from google.cloud import storage
from google.cloud import bigquery
from google.auth import default

from tmg.data import transfer
from tmg.data._helper import get_bq_write_disposition
from tmg.data import logs


class Client:
    """
    Client to bundle BigQuery functionality.

    Args:
        project (str): The Project ID for the project which the client acts on behalf of.
    """

    def __init__(self, project):
        self.project = project

        scopes = (
            'https://www.googleapis.com/auth/bigquery',
            'https://www.googleapis.com/auth/cloud-platform',
            'https://www.googleapis.com/auth/drive',
        )
        credentials, _ = default(scopes=scopes)
        self.bigquery_client = bigquery.Client(
            credentials=credentials,
            project=self.project
        )

    def download_table(self, table, local_folder='.', separator=',', print_header=True):
        """
        Export the table to the local file in CSV format.

        Args:
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            local_folder (:obj:`str`, Optional):  The local folder with out the slash at end. For example: ``/some_path/some_inside_path``. Defaults to current path :data:`.`
            separator (:obj:`str`, Optional): The separator. Defaults to :data:`,`
            print_header (:obj:`boolean`, optional):  True to print a header row in the exported file otherwise False. Defaults to :data:`True`.

        Returns:
            list: list of file names, if the table is big, multiple files are downloaded

        Examples:
            >>> from tmg.data import bq
            >>> client = bq.Client(project='my-project-id')
            >>> client.download_table(table='my-project-id.my_dataset.my_table')

        """

        storage_client = storage.Client(project=self.project)
        bucket_name = str(uuid.uuid4())
        logs.client.logger.info('Creating temporary bucket gs://{}'.format(bucket_name))
        tmp_bucket = storage_client.create_bucket(bucket_name, location='EU')

        transfer_client = transfer.Client(project=self.project)
        transfer_client.bq_to_gs(table, bucket_name, separator=separator, print_header=print_header)

        logs.client.logger.info('Getting the list of available blobs in gs://{}'.format(bucket_name))
        blobs = storage_client.list_blobs(bucket_name)
        blob_names = []
        for blob in blobs:
            logs.client.logger.info('Downloading gs://{}/{}'.format(bucket_name, blob.name))
            blob.download_to_filename('{}/{}'.format(local_folder, blob.name))
            blob_names.append(blob.name)
            logs.client.logger.info('Deleting gs://{}/{}'.format(bucket_name, blob.name))
            blob.delete()

        logs.client.logger.info('Deleting bucket gs://{}'.format(bucket_name))
        tmp_bucket.delete()

        return blob_names

    def upload_table(self, file_path, table, write_preference, separator=',', auto_detect=True, skip_leading_rows=True,
                     schema=(), partition_date=None, max_bad_records=0):
        """Import into the BigQuery table from the local file.

        Args:
            file_path (str):  The local file path
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            write_preference (str): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
            separator (str, Optional): The separator. Defaults to :data:`,`
            auto_detect (boolean, Optional):  True if the schema should automatically be detected otherwise False. Defaults to :data:`True`.
            skip_leading_rows (boolean, Optional):  True to skip the first row of the file otherwise False. Defaults to :data:`True`.
            schema (tuple, Optional): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field', 'STRING'))``. Defaults to ().
            partition_date (str, Optional): The ingestion date for partitioned destination table. For example: ``20210101``. The partition field name will be __PARTITIONTIME
            max_bad_records (int, Optional): The maximum number of invalid rows. Defaults to :data:`0`.

        Examples:
            >>> from tmg.data import bq
            >>> client = bq.Client(project='my-project-id')
            >>> client.upload_table(file_path='/my_path/my_filename', table='my-project-id.my-dataset.my-table')

        """

        project, dataset_id, table_id = table.split('.')
        dataset_ref = bigquery.DatasetReference(project=project, dataset_id=dataset_id)

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.CSV,
            skip_leading_rows=1 if skip_leading_rows else 0,
            autodetect=auto_detect if not schema else False,
            field_delimiter=separator,
            write_disposition=get_bq_write_disposition(write_preference),
            allow_quoted_newlines=True,
            max_bad_records=max_bad_records
        )

        if schema:
            job_config.schema = [bigquery.SchemaField(schema_field[0], schema_field[1]) for schema_field in schema]

        if partition_date:
            job_config.time_partitioning = bigquery.TimePartitioning(type_=bigquery.TimePartitioningType.DAY)
            table_id += '${}'.format(partition_date)

        table_ref = bigquery.TableReference(dataset_ref, table_id=table_id)

        with open(file_path, "rb") as source_file:
            logs.client.logger.info('Loading BigQuery table {} from file {}'.format(table, file_path))
            job = self.bigquery_client.load_table_from_file(source_file, table_ref, job_config=job_config)

        job.result()

    def run_query(self, query=None, query_file_name=None, params=(), destination=None, write_preference='empty',
                  partition_date=None):
        """Run the query and return the result

        Args:
            query (str) or query_file_name(str) is required.
            query (str, Optional): The query string.
            query_file_name (str, Optional): The query file path. For example: ``/my_query_path/my_query.sql``
                The use of variables in an SQL string can be done using ``{{var}}``.
                For example: ``SELECT first_name FROM profile where last_name={{last_name}}``
                More options of interactions, please read Jinja Documentation at
                https://jinja.palletsprojects.com/en/2.11.x/templates/
            params (dict, Optional): The query parameters. Each element should be {param_name: param_value}.
                For example: ``{"last_name", "Deniro", "age": 80}``
            destination (str, Optional): The BigQuery destination table name. For example: ``project.dataset.table``. Defaults to None
            write_preference (str, Optional): The option to specify what action to take when you load data from a source file. Value can be on of
                      ``'empty'``: Writes the data only if the table is empty.
                      ``'append'``: Appends the data to the end of the table.
                      ``'truncate'``: Erases all existing data in a table before writing the new data.
                      Defaults to :data:`empty`
            partition_date (str, Optional): The ingestion date for partitioned destination table. For example: ``20210101``. The partition field name will be __PARTITIONTIME

        Returns:
            The query results

        """
        if not (query or query_file_name):
            raise Exception("Parameter [query] or [query_file_name] must be provided")

        job_config = bigquery.QueryJobConfig(allow_large_results=True)
        if destination:
            if partition_date:
                job_config.time_partitioning = bigquery.TimePartitioning(type_=bigquery.TimePartitioningType.DAY)
                destination += '${}'.format(partition_date)
            job_config.destination = destination
            job_config.write_disposition = get_bq_write_disposition(write_preference)

        if query_file_name:
            with open(query_file_name, mode="r") as query_file:
                query = query_file.read()
        if params:
            query_template = Template(query)
            query = query_template.render(params)

        if destination:
            logs.client.logger.info(
                ('Writing query results into {destination}: \n' + 75 * '=' + '\n{query}\n' + 75 * '=').format(
                    destination=destination,
                    query=query
                )
            )
        else:
            logs.client.logger.info(('Running query: \n' + 75 * '=' + '\n{}\n' + 75 * '=').format(query))

        query_job = self.bigquery_client.query(query, job_config=job_config)
        result = query_job.result()
        logs.client.logger.info(
            'Total bytes processed: {:.2f} GiB'.format(query_job.total_bytes_processed/(1024*1024*1024))
        )

        return result

    def create_table(self, table, schema_file_name=None, schema_fields=None):
        """Create the BigQuery table

        Args:
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            schema_file_name (str, Optional): The schema file name. For example: ``/my_schema_path/schema.csv``.
                                              The schema file should be in CSV format and each line is FIELD_NAME, FIELD_TYPE, MODE. For example:
                                              ``id,STRING,REQUIRED``. Defaults to None
            schema_fields (tuple, Optional): The table schema.
                                             For example. ``(("id", "INT64", "REQUIRED"),("name", "STRING", "NULLABLE"))``. Defaults to None.

        """

        if not schema_fields and not schema_file_name:
            raise ValueError('Either schema file name or schema fields should be provided.')

        if schema_file_name:
            with open(schema_file_name) as schema_file:
                reader = csv.reader(schema_file)
                schema = [
                    bigquery.SchemaField(row[0], row[1], mode=row[2])
                    for row in reader
                ]
        else:
            schema = [
                bigquery.SchemaField(schema_field[0], schema_field[1], mode=schema_field[2])
                for schema_field in schema_fields
            ]

        table = bigquery.Table(table, schema=schema)
        self.bigquery_client.create_table(table)

    def delete_table(self, table):
        """Delete the BigQuery table

        Args:
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
        """
        self.bigquery_client.delete_table(table)

    def create_dataset(self, dataset, location='EU'):
        """Create the BigQuery dataset

        Args:
            dataset (str): The BigQuery dataset name. For example: ``my-project.my_dataset``.
            location (str, Optional): The location of the dataset. Defaults to EU.
        """
        dataset = bigquery.Dataset(f'{self.project}.{dataset}')
        dataset.location = location
        self.bigquery_client.create_dataset(dataset)

    def delete_dataset(self, dataset, delete_contents=False):
        """Delete the BigQuery dataset

        Args:
            dataset (str): The BigQuery dataset name. For example: ``my-project.my_dataset``.
            delete_contents (boolean, Optional): If True, delete all the tables in the dataset. If
                False and the dataset contains tables, the request will fail.
                Default is False.
        """
        self.bigquery_client.delete_dataset(dataset, delete_contents=delete_contents)




import logging


class Client:

    def __init__(self):
        """
        Logging configuration
        """

        self._logger = self._load_logger('tmg-data')

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, value):
        self._logger = value

    @staticmethod
    def _load_logger(name):
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)

        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(name)s - %(asctime)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        return logger


client = Client()



import parse
import paramiko
import os
import stat
import threading

from tmg.data import logs


class Client:
    """
    Creates a client which manages the connection to the FTP server.

    Args:
        connection_string (str): The FTP connection string in the format {username}:{password}@{host}:{port}
    """

    def __init__(self, connection_string, private_key=None):

        parsed_connection = parse.parse('{username}:{password}@{host}:{port}', connection_string)

        if private_key is None:
            ssh_client = paramiko.SSHClient()
            ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            ssh_client.connect(hostname=parsed_connection['host'],
                                username=parsed_connection['username'],
                                password=parsed_connection['password'],
                                port=int(parsed_connection['port']))
            self.connection = ssh_client.open_sftp()
        elif private_key and parsed_connection['password']:
            transport = paramiko.Transport((parsed_connection['host'],
                                            int(parsed_connection['port'])))
            transport.connect()

            key = paramiko.RSAKey.from_private_key_file(private_key)
            transport.auth_publickey(parsed_connection['username'], key)
            logs.client.logger.info('Authentication with private key successful')

            password_auth_handler = paramiko.auth_handler.AuthHandler(transport)
            transport.auth_handler = password_auth_handler
            transport.lock.acquire()

            password_auth_event = threading.Event()
            password_auth_handler.auth_event = password_auth_event
            password_auth_handler.auth_method = 'password'
            logs.client.logger.info('Authentication method set to password')

            password_auth_handler.username = parsed_connection['username']
            password_auth_handler.password = parsed_connection['password']

            userauth_message = paramiko.message.Message()
            userauth_message.add_string('ssh-userauth')
            userauth_message.rewind()
            password_auth_handler._parse_service_accept(userauth_message)
            transport.lock.release()
            password_auth_handler.wait_for_response(password_auth_event)

            self.connection = transport.open_sftp_client()

            logs.client.logger.info('SFTP connection opened successfully')

    def list_files(self, dir_path='.'):
        """Lists the files in a directory on the FTP server

        Args:
            dir_path (str): the path of the directory to list the contents of

        Returns:
            list: The list of the contents
        """

        return self.connection.listdir(dir_path)

    def upload_file(self, local_path, remote_path='.'):
        """Uploads a file from the local to the sFTP. If the remote path provided is a directory the filename
        of the local file will be used.

        Args:
            local_path (str): local path to the file to load
            remote_path (str): path to load the file to on the sFTP

        Returns:
            str: The path to the file loaded onto the sFTP
        """

        try:
            fileattr = self.connection.lstat(remote_path)
            if stat.S_ISDIR(fileattr.st_mode):
                filename = os.path.basename(local_path)
                remote_path = os.path.join(remote_path, filename)
        except IOError:
            pass

        self.connection.put(localpath=local_path, remotepath=remote_path)

        return remote_path

    def download_file(self, remote_path, local_path='.'):
        """Downloads a file from the sFTP to the local system. If the local path is a directory the filename of the
        remote file is used.

        Args:
            remote_path (str): The path to the file to download.
            local_path (str): The path to download the file to on the local system

        returns:
            str: The local path the file has been place.
        """

        if os.path.isdir(local_path):
            filename = os.path.basename(remote_path)
            local_path = os.path.join(local_path, filename)

        logs.client.logger.info("Starting file download.")

        self.connection.get(remotepath=remote_path, localpath=local_path)

        logs.client.logger.info("Finished downloading file.")

        return local_path

    def change_file_permission(self, remote_path, permission=stat.S_IRWXU):
        """Changes the permissions of the file on the sFTP

        Args:
            remote_path (str): The path to the file to change to permissions on
            permission (int): The permissions to set the file to
        """

        self.connection.chmod(path=remote_path, mode=permission)

    def delete_file(self, remote_path):
        """Deletes the specified file off the sFTP server

        Args:
            remote_path: path to the file to delete
        """

        self.connection.remove(remote_path)

    def close(self):
        """Closes the connection to the ftp server
        """
        self.connection.close()



import unittest
import csv
import time
import mysql.connector
from google.cloud import bigquery
from google.cloud import storage

from tests.setup import setup
from tmg.data import runner, transfer


def setUpModule():
    setup.create_bq_table()
    setup.create_bucket()
    setup.create_mysql_db()
    setup.create_mysql_table()


def tearDownModule():
    setup.delete_bq_dataset()
    setup.delete_bucket()
    setup.delete_mysql_db()


class TestRunner(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestRunner, self).__init__(*args, **kwargs)
        self.setup = setup

    def test_add(self):

        def test(a, b, c=1):
            pass

        parallel_runner = runner.ParallelRunner()
        parallel_runner.add("test1", test, "a", "b", c=2)

        name, function, args, kwargs = parallel_runner.processes[0]

        self.assertEqual(function.__name__, "test")
        self.assertEqual(args, ["a", "b"])
        self.assertDictEqual(kwargs, {"c": 2})

    def test_run(self):

        def test_function(run_name, sleep_length=5):
            print(f"starting run {run_name}")
            time.sleep(sleep_length)
            print(f"finished run {run_name}")

        parallel_runner = runner.ParallelRunner()
        parallel_runner.add("test1", test_function, "run_1")
        parallel_runner.add("test2", test_function, "run_2", sleep_length=10)

        parallel_runner.run(wait_for_result=True)

    def test_run_integration(self):

        parallel_runner = runner.ParallelRunner()

        # First transfer
        bq_gs_client = transfer.Client(project=self.setup.project)
        parallel_runner.add(
            "bq_to_gs",
            bq_gs_client.bq_to_gs,
            table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'actors'),
            bucket_name=self.setup.bucket_name
        )

        # Second Transfer
        mysql_bq_client = transfer.Client(project=self.setup.project)
        parallel_runner.add(
            "mysql_to_bq",
            mysql_bq_client.mysql_to_bq,
            instance_name=self.setup.mysql_instance_name,
            database=self.setup.mysql_database,
            query='SELECT * from actors',
            bq_table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'actors_from_mysql'),
            bq_table_schema=(('profile_id', 'STRING'), ('first_name', 'STRING'), ('last_name', 'STRING')),
            write_preference='truncate'
        )

        # Run both transfers in parallel
        results = parallel_runner.run(wait_for_result=True)
        file_names = results[0]

        ### Getting results of first transfer
        # creating list based on source big query table values
        bq_gs_bq_keys = []
        bq_gs_bigquery_client = bigquery.Client(project=self.setup.project)
        job = bq_gs_bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                'actors'
            )
        )
        for row in job.result():
            bq_gs_bq_keys.append('{}{}{}'.format(
                row.profile_id,
                '' if row.first_name is None else row.first_name,
                '' if row.last_name is None else row.last_name,
            ))

        # creating list based on destination bucket file values
        bq_gs_storage_client = storage.Client(project=self.setup.project)
        with open('actors.csv', 'wb') as file_obj:
            bq_gs_storage_client.download_blob_to_file(blob_or_uri=file_names[0], file_obj=file_obj)

        with open('actors.csv', newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            bq_gs_storage_keys = []
            for row in reader:
                bq_gs_storage_keys.append(
                    str(row['profile_id']) + row['first_name'] + row['last_name']
                )

        ###Getting results based on second transfer
        # creating list based on source mysql table values
        connection = mysql.connector.connect(
            host=self.setup.mysql_ip_address,
            user=self.setup.mysql_username,
            password=self.setup.mysql_password,
            database=self.setup.mysql_database
        )
        cursor = connection.cursor()
        cursor.execute('SELECT profile_id, first_name, last_name from actors')
        mysql_bq_mysql_keys = []
        for row in cursor.fetchall():
            mysql_bq_mysql_keys.append('{}{}{}'.format(
                row[0],
                'NULL' if row[1] is None else row[1],
                'NULL' if row[2] is None else row[2]
            ))
        cursor.close()
        connection.close()

        # creating list based on destination bucket values
        mysql_bq_bq_keys = []
        mysql_bq_bigquery_client = bigquery.Client(project=self.setup.project)
        job = mysql_bq_bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                'actors_from_mysql'
            )
        )
        for row in job.result():
            mysql_bq_bq_keys.append(
                str(row.profile_id) +
                ('NULL' if row.first_name is None else row.first_name) +
                ('NULL' if row.last_name is None else row.last_name)
            )

        self.assertListEqual(mysql_bq_mysql_keys, mysql_bq_bq_keys)




import os
import io
import setuptools


name = "tmg-data"
description = "TMG data library"
version = "0.1.17"
dependencies = [
    "oauth2client==4.1.3",
    "google-api-python-client==1.12.8",
    "google-cloud-bigquery==1.24.0",
    "google-cloud-storage==1.27.0",
    "paramiko==2.7.1",
    "Jinja2==2.11.2",
    "mysql-connector==2.2.9",
    "boto3==1.14.8",
    "simple-salesforce==1.10.1",
    "parse==1.15.0",
    "delegator.py==0.1.1"
]

package_root = os.path.abspath(os.path.dirname(__file__))

readme_filename = os.path.join(package_root, "README.rst")
with io.open(readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read()


setuptools.setup(
    name=name,
    version=version,
    description=description,
    long_description=readme,
    author='TMG Data Platform team',
    author_email="data.platform@telegraph.co.uk",
    license="Apache 2.0",
    url='https://github.com/telegraph/tmg-data',
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    packages=setuptools.find_packages(),
    install_requires=dependencies,
    python_requires='>=3.6',

)



import unittest
import paramiko
import os
import stat

from tmg.data.ftp import Client
from tests.setup import setup


def setUpModule():
    setup.start_sftp_instance()
    setup.upload_ftp_files()


def tearDownModule():
    setup.remove_ftp_files()
    setup.stop_sftp_instance()


class TestFTP(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestFTP, self).__init__(*args, **kwargs)
        self.setup = setup

    def test_list_files(self):
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(hostname=self.setup.ftp_hostname,
                           username=self.setup.ftp_username,
                           password=self.setup.ftp_password,
                           port=self.setup.ftp_port)
        connection = ssh_client.open_sftp()

        test_client = Client(
            connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}"
        )

        self.assertListEqual(
            connection.listdir(),
            test_client.list_files()
        )

        self.assertListEqual(
            connection.listdir('test_dir/test_subdir'),
            test_client.list_files('test_dir/test_subdir')

        )

    def test_upload_file(self):
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(hostname=self.setup.ftp_hostname,
                           username=self.setup.ftp_username,
                           password=self.setup.ftp_password,
                           port=self.setup.ftp_port)
        connection = ssh_client.open_sftp()

        test_client = Client(
            connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}"
        )

        test_client.upload_file('tests/data/sample.csv')
        self.assertTrue(connection.lstat('sample.csv'))

        test_client.upload_file('tests/data/sample.csv', remote_path='test_dir')
        self.assertTrue(connection.lstat('test_dir/sample.csv'))

        test_client.upload_file('tests/data/sample.csv',
                                remote_path='test_dir/test_subdir/rename_sample.csv')
        self.assertTrue(connection.lstat('test_dir/test_subdir/rename_sample.csv'))

    def test_download_file(self):
        test_client = Client(
            connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}"
        )

        test_client.download_file('test_dir/download_sample.csv')
        self.assertTrue(os.path.exists('download_sample.csv'))

        test_client.download_file('test_dir/download_sample.csv', local_path='tests/data')
        self.assertTrue(os.path.exists('tests/data/download_sample.csv'))

        test_client.download_file('test_dir/download_sample.csv', local_path='tests/data/rename_sample.csv')
        self.assertTrue(os.path.exists('tests/data/rename_sample.csv'))

    def test_change_file_permission(self):
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(hostname=self.setup.ftp_hostname,
                           username=self.setup.ftp_username,
                           password=self.setup.ftp_password,
                           port=self.setup.ftp_port)
        connection = ssh_client.open_sftp()

        test_client = Client(
            connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}"
        )

        test_client.change_file_permission('test_dir/permission_change.csv')
        self.assertEqual(oct(connection.stat('test_dir/permission_change.csv').st_mode)[-3:], '700')

        test_client.change_file_permission('test_dir/permission_change2.csv',
                                           stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH)
        self.assertEqual(oct(connection.stat('test_dir/permission_change2.csv').st_mode)[-3:], '754')

    def test_delete_file(self):
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(hostname=self.setup.ftp_hostname,
                           username=self.setup.ftp_username,
                           password=self.setup.ftp_password,
                           port=self.setup.ftp_port)
        connection = ssh_client.open_sftp()

        test_client = Client(
            connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}"
        )

        test_client.delete_file('test_dir/delete_sample.csv')

        with self.assertRaises(IOError):
            connection.stat('test_dir/delete_sample.csv')

    def test_close(self):
        test_client = Client(
            connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}"
        )

        self.assertFalse(test_client.close())
